#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
## `velox.exceptions`

The `velox.exceptions` submodule defines all the exceptions that are specific 
to the Velox ecosystem.
"""


class VeloxCreationError(Exception):
    """
    Raised in the event of a class instantiation that does not follow
    protocol.
    """
    pass


class VeloxConstraintError(Exception):
    """
    Raised in the event of no matches for loading a model
    """
    pass



import pytest

import os
from glob import glob
from backports.tempfile import TemporaryDirectory
import pickle
import time
import numpy as np


from velox.filesystem import (get_aware_filepath, find_matching_files,
                              stitch_filename, ensure_exists, parse_s3)

from velox.tools import timestamp, obtain_padding_bytes

import boto3
from moto import mock_s3

import logging
logging.basicConfig(level=logging.DEBUG)

TEST_BUCKET = 'ci-velox-bucket'

# TODO(lukedeo): Refactor the test suite to be more fixtured-out to make
# sure we don't have duplicated code.


@mock_s3
def test_aware_filepath():
    conn = boto3.resource('s3', region_name='us-east-1')
    # We need to create the bucket since this is all in Moto's 'virtual' AWS
    # account
    conn.create_bucket(Bucket=TEST_BUCKET)

    with TemporaryDirectory() as d:
        fpaths = ['s3://{}/file.txt'.format(TEST_BUCKET),
                  os.path.join(d, 'file.txt')]

        for fp in fpaths:
            with pytest.raises(ValueError):
                with get_aware_filepath(fp, 'w+') as f:
                    f.write('foobar')
            with get_aware_filepath(fp, 'w') as f:
                f.write('foobar')

            with get_aware_filepath(fp, 'r') as f:
                assert f.read() == 'foobar'


@mock_s3
def test_aware_filepath_with_type_hint():
    conn = boto3.resource('s3', region_name='us-east-1')
    # We need to create the bucket since this is all in Moto's 'virtual' AWS
    # account
    conn.create_bucket(Bucket=TEST_BUCKET)
    x = np.random.normal(0, 1, (10, ))
    with TemporaryDirectory() as d:
        fpaths = [
            's3://{}/file.txt'.format(TEST_BUCKET),
            os.path.join(d, 'file.txt')
        ]

        for fp in fpaths:
            with get_aware_filepath(fp, 'wb') as f:
                f.write(b'foobarbaz')
                f.write(obtain_padding_bytes(x))

            with get_aware_filepath(fp, 'rb') as f:
                assert f.read() == b'foobarbaz'

            with get_aware_filepath(fp, 'rb', yield_type_hint=True) as \
                    (f, type_hint):
                assert f.read() == b'foobarbaz'
                assert type_hint == 'numpy.ndarray'


def test_stitch_filename():

    reference = 's3://myBucket/file.txt'

    assert stitch_filename('s3://myBucket/', '/file.txt') == reference
    assert stitch_filename('s3://myBucket', '/file.txt') == reference
    assert stitch_filename('s3://myBucket', 'file.txt') == reference
    assert stitch_filename('s3://myBucket/', 'file.txt') == reference

    with TemporaryDirectory() as d:
        reference = os.path.join(d, 'file.txt')
        assert stitch_filename(d, 'file.txt') == reference


@mock_s3
def test_ensure_exists():
    ensure_exists('s3://{}'.format(timestamp()))
    with TemporaryDirectory() as d:
        ensure_exists(os.path.join(d, timestamp()))


def test_bad_s3():
    with pytest.raises(ValueError):
        with TemporaryDirectory() as d:
            _ = parse_s3(d)


@mock_s3
def test_list_contents():
    conn = boto3.resource('s3', region_name='us-east-1')
    # We need to create the bucket since this is all in Moto's 'virtual' AWS
    # account
    conn.create_bucket(Bucket=TEST_BUCKET)

    nb_files = 5

    with TemporaryDirectory() as d:

        fpaths = ['s3://{}'.format(TEST_BUCKET) + '/file-{}.txt',
                  os.path.join(d, 'file-{}.txt')]

        for fp in fpaths:
            for i in range(nb_files):
                with get_aware_filepath(fp.format(i), 'w') as f:
                    f.write('foobar-{}'.format(i))

        s3match = find_matching_files('s3://{}'.format(TEST_BUCKET),
                                      specifier='*txt')
        fsmatch = find_matching_files(d, specifier='*txt')

        assert len(s3match) == nb_files
        assert len(fsmatch) == nb_files



import pytest

from backports.tempfile import TemporaryDirectory

import boto3
from moto import mock_s3
import numpy as np
from velox.wrapper import SimplePickle, SimpleKeras
from velox.obj import load_velox_object
from keras.layers import Dense
from keras.models import Sequential


def build_test_data():

    X = np.random.normal(0, 1, (100, 3))
    X_test = np.random.normal(0, 1, (100, 3))
    y = X.dot([1, 2, 3])
    y += np.random.normal(0, 5, y.shape)

    return X, X_test, y


def test_simplepickle():
    from sklearn.linear_model import SGDRegressor

    X, X_test, y = build_test_data()

    clf = SGDRegressor(n_iter=20).fit(X, y)

    y_orig = clf.predict(X_test)

    with TemporaryDirectory() as d:
        SimplePickle(clf).save(prefix=d)
        model = SimplePickle.load(prefix=d)

    assert np.allclose(y_orig, model.predict(X_test))


@mock_s3
def test_load_function_with_wrapper_class_local_and_s3():
    from sklearn.linear_model import SGDRegressor

    X, X_test, y = build_test_data()
    clf = SGDRegressor(n_iter=20).fit(X, y)
    y_orig = clf.predict(X_test)

    conn = boto3.resource('s3', region_name='us-east-1')
    # We need to create the bucket since this is all in Moto's 'virtual' AWS
    # account
    TEST_BUCKET = 'foo-test-bucket'
    conn.create_bucket(Bucket=TEST_BUCKET)

    with TemporaryDirectory() as tmpdir:
        for d in [tmpdir, 's3://{}/bazprefix/'.format(TEST_BUCKET)]:
            SimplePickle(clf).save(prefix=d)
            model = load_velox_object('simplepickle', prefix=d)
            assert np.allclose(y_orig, model.predict(X_test))


def test_simplekeras():

    X, X_test, y = build_test_data()

    net = Sequential([
        Dense(20, activation='relu', input_dim=3),
        Dense(1)
    ])

    net.compile('adam', 'mse')

    net.fit(X, y, epochs=1)

    y_orig = net.predict(X_test)

    with TemporaryDirectory() as d:
        SimpleKeras(net).save(prefix=d)
        model = SimpleKeras.load(prefix=d)

    assert np.allclose(y_orig, model.predict(X_test))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
## `velox.lite`

The `velox.lite` submodule provides a lightweight version of Velox's binary 
management capabilities.

The core functionality is geared towards continuous deployment environments in
the machine learning world, where consistency and versioning of binary objects
is key to maintain system integrity.

Suppose we build a simple [`scikit-learn`](http://scikit-learn.org/) model that we want to be available somewhere else in a secure, verifyable manner. 

Suppose a data scientist trains the following model.
<!--begin_code-->

    #!python
    import os

    from sklearn.linear_model import SGDClassifier
    from sklearn.datasets import make_blobs
    from velox.lite import save_object

    X, y = make_blobs()

    clf = SGDClassifier()
    clf.fit(X, y)

    save_object(
        obj=clf, 
        name='CustomerModel', 
        prefix='s3://myprodbucket/ml/models', 
        secret=os.environ.get('ML_MODELS_SECRET')
    )
<!--end_code-->

Elsewhere, on a production server, we could easily load this model and
verify it's integrity.

<!--begin_code-->

    #!python
    import os

    from velox.lite import load_object
    try:
        clf = load_object(
            name='CustomerModel', 
            prefix='s3://myprodbucket/ml/models', 
            secret=os.environ.get('ML_MODELS_SECRET')
        )
    except RuntimeError:
        raise RuntimeError('Invalid Secret!')

    # do things with clf...
<!--end_code-->
"""
import dill
import io
import logging

import itsdangerous
from semantic_version import Version as SemVer, Spec as Specification

from . import exceptions
from . import filesystem
from . import tools


DEFAULT_SECRET = 'velox'


logger = logging.getLogger(__name__)


def _get_deserialization_hook(classname):
    if classname == 'dill':
        return dill.load
    else:
        classdef = tools.import_from_qualified_name(classname)
        return classdef._load


def _get_serialization_hook(obj):
    if hasattr(obj, '_save'):
        if not callable(obj._save):
            raise TypeError(
                '_save attribute on object of type {} is not callable.'
                .format(type(obj))
            )
        serialization_hook = obj._save
        deserialization_class = tools.fullname(obj)
    # TODO(lukedeo): create multiple cases here where we can handle custom
    # types like Keras models or PyTorch modules.
    else:
        serialization_hook = lambda buf: dill.dump(obj, buf)
        # We don't need to know the object class here, we just use dill
        deserialization_class = 'dill'

    return serialization_hook, deserialization_class


def _seems_like_nonlite_velox_obj(prefix, name):
    """
    Check and see if the prefix / name combo could match a subclassed velox
    object
    """
    specifier = '*_{}*.vx'.format(name)
    if filesystem.find_matching_files(prefix, specifier):
        return True
    return False


def save_object(obj, name, prefix, versioned=False, secret=None, bump='patch'):
    """
    Velox-managed method to save generic Python objects. Affords the ability
    to version saved objects to a common prefix, as well as to sign binaries
    with a secret.

    * If `obj` has a callable method called `_save`, it will call the method
        to save the object with the signature `obj._save(buf)` where buf is an
        object of type `io.BytesIO`.

    * Else, will use the fantastic `dill` library to save the 
        object generically.

    Saving / loading can proceed by two different standards - the first is a 
    versioned saving / loading scheme, where objects are saved according 
    to `/path/to/prefix/objectname-v1`, and the second is unversioned where 
    objects are saved according to `/path/to/prefix/objectname`. Versions are 
    automatically assigned - they autoincrement whenever a new object is pushed
    to the given prefix.

    Args:
    -----

    * `obj (object)`: Object that is either pickle-able / dill-able or 
        defines a `_save(...)` method to serialize to a `io.BytesIO` object.

    * `name (str)`: The name to save the object under at the `prefix` location.

    * `prefix (str)`: the prefix (can be on S3 or on a local filesystem) to
        save the managed object to. If on S3, it is expected to take the form
        `s3://my-bucket/other/things/`.

    * `versioned (bool)`: Whether or not to save the object according to a 
        versioned scheme.

    * `secret (str)`: A secret (**guard like a password**) to require in 
        the deserialization process.

    * `bump (str)`: One of `{major, minor, patch}`, indicates the semantic
        version bump to save the `obj` with. Consult with the 
        [semantic versioning website](https://semver.org/) for more information.

    Returns:
    --------

    `filename`: the final filename that `obj` is saved into. 


    Raises:
    -------

    * `IOError` if attempting to save an unversioned file that already exists.

    * `ValueError` if a semantic version string cannot be parsed.
    """
    serialization_hook, deserialization_class = _get_serialization_hook(obj)

    # Managed objects can either be versioned or unversioned - in the
    # versioned case, they will always have the name
    # `/my/prefix/myservable-v0.2.3` where vX will be a semver string. If not
    # versioned, then will simply be `/my/prefix/myservable`
    if versioned:
        matching_files = filesystem.find_matching_files(
            prefix=prefix,
            specifier='{}-v*'.format(name)
        )

        # We first try to parse all the substrings defined by the last RHS of a
        # block seperated by a `-v`.
        try:
            matched_versions = [SemVer(f.split('-v')[-1])
                                for f in matching_files]
        except ValueError as err:
            raise ValueError('Error parsing semantic version string: {}'
                             .format(err))

        if matched_versions:
            latest_version = max(matched_versions)
            if bump == 'patch':
                version = latest_version.next_patch()
            elif bump == 'minor':
                version = latest_version.next_minor()
            elif bump == 'major':
                version = latest_version.next_major()
            else:
                raise ValueError('invalid bump: {}'.format(bump))
        else:
            version = SemVer('0.1.0')

        logger.debug('assigning version v{}'.format(version))
        filename = filesystem.stitch_filename(prefix,
                                              '{}-v{}'.format(name, version))

    else:
        filename = filesystem.stitch_filename(prefix, name)
        matching_files = filesystem.find_matching_files(
            prefix=prefix,
            specifier=name
        )
        if matching_files:
            raise IOError('File: {} already exists'.format(filename))

    logger.debug('will use filename: {} for serialization'.format(filename))
    filesystem.ensure_exists(prefix)
    buf = io.BytesIO()
    serialization_hook(buf)
    buf = buf.getvalue()

    data = {
        'data': buf,
        'class': deserialization_class
    }

    serializer = itsdangerous.Serializer(secret or DEFAULT_SECRET,
                                         serializer=dill)
    if secret:
        logger.debug('specified SECRET=<{}>'.format('*' * len(secret)))
    with filesystem.get_aware_filepath(filename, 'wb') as fileobject:
        serializer.dump(data, fileobject)

    return filename


def load_object(name, prefix, versioned=False, version=None, secret=None,
                return_sha=False):
    """
    Velox-managed method to load generic Python objects that have been saved
    via `velox.lite.save_object`. Affords the ability to load versioned
    or unversioned saved objects from a common prefix, as well as to do basic
    signature verification with a secret.

    * If the object type that was saved has a `classmethod` called `_load`,
        it will call the method to load the object with the
        signature `ObjectClass._load(buf)` where buf is an object of
        type `io.BytesIO`. The `ObjectClass` is determined from a small piece
        of data written by Velox in the `velox.lite.save_object` method.

    * Else, will use the `dill` library to load the object.

    Args:
    -----

    * `name (str)`: The name of the object to load.

    * `prefix (str)`: the prefix (can be on S3 or on a local filesystem) to
        load the managed object from. If on S3, it is expected to take the form
        `s3://my-bucket/other/things/`.

    * `versioned (bool)`: Whether or not to load the object according to a 
        versioned scheme.

    * `version (str)`: A specific object semantic version to search with, or a
        semver compatible version specification (such as `<=1.0.2,>0.9.1`).

    * `secret (str)`: A secret (**guard like a password**) to verify 
        permissions in the deserialization process.

    * `return_sha (bool)`: Whether or not to return the sha as part of the 
        payload. If True, returns (obj, sha), else, just returns obj.

    Returns:
    --------

    Returns the loaded object `obj` if `not return_sha`, else, will return a
    tuple of `(obj, sha)`    

    Raises:
    -------

    * `velox.exceptions.VeloxConstraintError` if no matching files to load the
        object from are found.

    * `velox.exceptions.RuntimeError` if `secret` does not match the secret
        that was used to save the object or if a pinned version load is 
        attempted with an unversioned loading scheme.
    """
    if version and not versioned:
        raise RuntimeError('Cannot perform a search against a specific '
                           'version with unversioned loading scheme')
    try:
        if versioned:
            matching_files = filesystem.find_matching_files(
                prefix=prefix,
                specifier='{}-v*'.format(name)
            )
            if not matching_files:
                raise exceptions.VeloxConstraintError(
                    'No matching files at prefix: {} with name: {}. '
                    'Did you mean to load this binary with '
                    'an unversioned scheme?'
                    .format(prefix, name)
                )
            matched_versions = [SemVer(f.split('-v')[-1])
                                for f in matching_files]

            if version:
                best_version = Specification(version).select(matched_versions)
                if not best_version:
                    raise exceptions.VeloxConstraintError(
                        'No matching files at prefix: {} with '
                        'name: {} and version: {}. '
                        .format(prefix, name, version)
                    )
            else:
                best_version = max(matched_versions)
            filename = filesystem.stitch_filename(prefix,
                                                  '{}-v{}'
                                                  .format(name, best_version))
        else:
            filename = filesystem.stitch_filename(prefix, name)
            matching_files = filesystem.find_matching_files(
                prefix=prefix,
                specifier=name
            )
            if not matching_files:
                raise exceptions.VeloxConstraintError(
                    'No matching files at prefix: {} with name: {}. '
                    'Did you mean to load this binary with a versioned scheme?'
                    .format(prefix, name)
                )
    # If someone is trying to use this method to load an object from a
    # non-lite version of velox, let's catch that.
    except exceptions.VeloxConstraintError as err:
        if _seems_like_nonlite_velox_obj(prefix, name):
            logger.info('Found a saved object that resembles a subclassed & '
                        'managed object. Reverting to load_velox_object')
            from .obj import load_velox_object
            return load_velox_object(registered_name=name, prefix=prefix,
                                     version_constraints=version)
        raise err

    logger.debug('will load from filename: {}'.format(filename))
    serializer = itsdangerous.Serializer(secret or DEFAULT_SECRET,
                                         serializer=dill)

    sha = None
    with filesystem.get_aware_filepath(filename, 'rb') as fileobject:
        try:
            data = serializer.load(fileobject)
        except itsdangerous.BadSignature:
            raise RuntimeError(
                'Mismatched secret - deserialization not authorized'
            )
        sha = tools.sha(data['data'])
        deserialization_hook = _get_deserialization_hook(data['class'])
        obj = deserialization_hook(io.BytesIO(data['data']))

    return (obj, sha) if return_sha else obj



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
## `velox.wrapper`

The `velox.wrapper` submodule provides pre-packaged wrappers around Keras
models and around generally pickleable objects in the python ecosystem.
"""

import dill

from .obj import VeloxObject, register_object, _fail_bad_init, _zero_downtime


@register_object(registered_name='simplepickle')
class SimplePickle(VeloxObject):
    """ SimplePickle is a passthru wrapper for any pickle-able Python object,
    allowing you to save, load, and swap in a consistent manner.

    Args:
    -----

    * `managed_object (object)`: an pickleable object


    Example:
    ---------

        #!python
        managed_object = SimplePickle({'foo': 'bar'}})

        print(managed_object.save(prefix='/path/to/saved/model'))

        # accesses underlying dictionary attribute.
        managed_object.get('foo')
    """

    def __init__(self, managed_object=None):
        super(SimplePickle, self).__init__()
        self._managed_object = managed_object

    def _save(self, fileobject):
        dill.dump(self, fileobject)

    @classmethod
    def _load(cls, fileobject):
        return dill.load(fileobject)

    @_fail_bad_init
    @_zero_downtime
    def __getattr__(self, name):
        try:
            return VeloxObject.__getattr__(self, name)
        except AttributeError:
            return getattr(self._managed_object, name)


@register_object(registered_name='simplekeras')
class SimpleKeras(VeloxObject):
    """ SimpleKeras is a passthru wrapper for a keras model, allowing you to
    save, load, and swap in a consistent manner.

    Args:
    -----

    * `keras_model(keras.models.Model)`: an instance of a Keras model

    Raises:
    -------

    * `TypeError` if keras_model is not None and is not a Keras model

    Notes:
    ------

    a user can access attributes of the underlying object by simply the
    attribute normally.

    Example:
    ---------

        #!python
        net = keras.models.Model(x, y)
        managed_object = SimpleKeras(net)

        # will use the Velox save rather than the Keras model save
        print(managed_object.save(prefix='/path/to/saved/model'))

        # accesses underlying keras attribute.
        managed_object.summary()

    """

    def __init__(self, keras_model=None):

        from keras.models import Model
        if keras_model is not None and not isinstance(keras_model, Model):
            raise TypeError('must be a Keras model - found type: {}'
                            .format(type(keras_model)))

        super(SimpleKeras, self).__init__()
        self._keras_model = keras_model

    def _save(self, fileobject):
        self._keras_model.save(fileobject.name, include_optimizer=False)

    @classmethod
    def _load(cls, fileobject):
        from keras.models import load_model
        o = cls()
        setattr(o, '_keras_model', load_model(fileobject.name))
        return o

    @_fail_bad_init
    @_zero_downtime
    def __getattr__(self, name):
        try:
            return VeloxObject.__getattr__(self, name)
        except AttributeError:
            return getattr(self._keras_model, name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
## `velox.tools`

The `velox.tools` submodule provides general support utilities to the Velox 
package.
"""

from __future__ import unicode_literals

import datetime
from hashlib import sha1
import six
from threading import Thread
import importlib
from concurrent.futures import Future

from builtins import bytes

VELOX_NEW_FILE_SIGNATURE = '||vx||'
VELOX_NEW_FILE_META_LENGTH = 100
VELOX_NEW_FILE_PAD_CHAR = '%'

VELOX_NEW_FILE_SIGNATURE_LENGTH = len(VELOX_NEW_FILE_SIGNATURE)
VELOX_NEW_FILE_FORMAT_STRING = (
    '{:' + VELOX_NEW_FILE_PAD_CHAR + '>' + str(VELOX_NEW_FILE_META_LENGTH) +
    '}' + VELOX_NEW_FILE_SIGNATURE
)
VELOX_NEW_FILE_EXTRAS_LENGTH = len(VELOX_NEW_FILE_FORMAT_STRING.format(''))


def sha(s):
    """
    get a simple, potentially value-inconsistent SHA1 of a python object.
    """
    m = sha1()
    if isinstance(s, six.string_types):
        m.update(bytes(s, 'utf-8'))
    elif isinstance(s, bytes):
        m.update(s)
    else:
        m.update(s.__repr__())
    return m.hexdigest()


def timestamp():
    """ 
    Returns a string of the form YYYYMMDDHHmmSSCCCCCC, where 

    * `YYYY` is the current year
    * `MM` is the current month, with zero padding to the left
    * `DD` is the current day, zero padded on the left
    * `HH`is the current hour, zero padded on the left
    * `mm`is the current minute, zero padded on the left
    * `CCCCCC`is the current microsecond as a decimal number, 
        zero padded on the left
    """
    return datetime.datetime.utcnow().strftime('%Y%m%d%H%M%S%f')


class abstractclassmethod(classmethod):

    # this hack allows us to enforce the ABC implementation of a classmethod
    __isabstractmethod__ = True

    def __init__(self, callable):
        callable.__isabstractmethod__ = True
        super(abstractclassmethod, self).__init__(callable)


def call_with_future(fn, future, args, kwargs):
    try:
        result = fn(*args, **kwargs)
        future.set_result(result)
    except Exception as exc:
        future.set_exception(exc)


def fullname(o):
    module = o.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return o.__class__.__name__
    return module + '.' + o.__class__.__name__


# def import_from_qualified_name(name):
#     components = name.split('.')
#     mod = __import__(components[0])
#     for comp in components[1:]:
#         mod = getattr(mod, comp)
#     return mod


def import_from_qualified_name(name):
    components = name.split('.')
    module, classname = components[:-1], components[-1]
    module = '.'.join(module)
    return getattr(importlib.import_module(module), classname)


def obtain_padding_bytes(obj, asbytes=True):
    qualname = fullname(obj)
    if len(qualname) > VELOX_NEW_FILE_META_LENGTH:
        raise ValueError('Qualified name {} is too long for use in Velox.'
                         .format(qualname))

    pad_bytes = VELOX_NEW_FILE_FORMAT_STRING.format(fullname(obj))

    if asbytes:
        return pad_bytes.encode()
    return pad_bytes


def obtain_qualified_name(meta_string):
    # No need to return the file signature
    if isinstance(meta_string, bytes):
        meta_string = meta_string.decode()
    class_info_string = meta_string[:-VELOX_NEW_FILE_SIGNATURE_LENGTH]
    return class_info_string.replace(VELOX_NEW_FILE_PAD_CHAR, '')


def get_file_meta(filehandle, truncate=False):
    """Get the metadata string from a velox file that has a type hint."""
    # Go to the end of the file
    filehandle.seek(0, 2)

    # We'll use the size to backtrack a few chars
    size = filehandle.tell()
    if size <= VELOX_NEW_FILE_SIGNATURE_LENGTH:
        return None
    filehandle.seek(size - VELOX_NEW_FILE_SIGNATURE_LENGTH)
    file_signature = filehandle.read()

    if isinstance(file_signature, bytes):
        file_signature = file_signature.decode()

    # If we're dealing with the old file format, we cannot find the qualname
    if file_signature != VELOX_NEW_FILE_SIGNATURE:
        return None

    # Go to the beginning of the file meta section
    filehandle.seek(size - VELOX_NEW_FILE_EXTRAS_LENGTH)

    meta_string = filehandle.read()

    if truncate:
        filehandle.seek(size - VELOX_NEW_FILE_EXTRAS_LENGTH)
        filehandle.truncate()

    filehandle.seek(0)

    return meta_string


def threaded(fn):
    """
    A simple decorator that allows a function to be called with its return 
    value given as a `Future` object.

    Example:
    --------

        #!python
        @threaded
        def fn(x):
            return x ** 2

        a = fn(6)
        assert a.result() == 36
    """

    def wrapper(*args, **kwargs):
        future = Future()
        Thread(target=call_with_future,
               args=(fn, future, args, kwargs)).start()
        return future
    return wrapper

__all__ = ['threaded', 'timestamp']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
file: setup.py
description: setuptools for Velox
author: Luke de Oliveira (lukedeo@vaitech.io)
"""

import os
from setuptools import setup
from setuptools import find_packages


from setuptools import setup, find_packages

VERSION = '0.5.0'

setup(
    name='Velox',
    version=VERSION,
    description=('Batteries-included tooling for handling promotion, '
                 'versioning, and zero-downtime requirments of Machine '
                 'Learning models in production.'),
    author='Luke de Oliveira',
    author_email='lukedeo@vaitech.io',
    url='https://github.com/vaitech/Velox',
    download_url='https://github.com/vaitech/Velox/archive/{}.tar.gz'.format(
        VERSION),
    license='Apache 2.0',
    install_requires=['apscheduler', 'semantic_version', 'dill',
                      'six', 'itsdangerous', 'futures;python_version=="2.7"',
                      'future;python_version=="2.7"'],
    packages=find_packages(),
    keywords=('Machine-Learning TensorFlow Deployment Versioning Keras '
              'AWS Deep Learning'),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6'
    ],
    extras_require={
        'aws': ['boto3'],
        'tests': ['numpy', 'pytest', 'pytest-cov', 'pytest-pep8',
                  'pytest-xdist', 'python-coveralls', 'moto', 'keras[h5py]',
                  'backports.tempfile', 'scikit-learn', 'mock',
                  'tensorflow<1.5'],
        'docs': ['bs4', 'strif', 'pdoc']
    }
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

u"""

<img src="https://github.com/vaitech/Velox/raw/master/img/velox-logo.png" width=37% align="right" />

# Welcome to Velox!

[![Build Status](https://travis-ci.org/vaitech/Velox.svg?branch=master)](https://travis-ci.org/vaitech/Velox)
[![Coverage Status](https://coveralls.io/repos/github/vaitech/Velox/badge.svg)](https://coveralls.io/github/vaitech/Velox)
[![PyPI version](https://badge.fury.io/py/Velox.svg)](https://badge.fury.io/py/Velox)

Please join us on [GitHub](https://github.com/vaitech/Velox)!

Deploying and managing live machine learning models is difficult. It involves a mix of handling model versioning, hot-swapping new versions and determining version constraint satisfaction on-the-fly, and managing binary file movement either on a networked or local file system or with a cloud storage system like S3. Velox can handle this for you with a simple base class enforcing opinionated methods of handling the above problems.

Velox provides three main utilities:

* Velox abstracts the messiness of consistent naming schemes and handling saving and loading requirements for a filesystem and for other forms of storage.
* Velox allows the ability to do a model / blob hotswap in-place for a new binary version.
* A single entry point (`velox.obj.load_velox_object`) to be able to load any object in any environment in a completely parameterized, runtime-dependent manner.

## Requirements
---

**Velox supports Python 2 and Python 3**

The main requirements are `apscheduler` for scheduling hot-swaps, `semantic_version` for version sanity, and the `futures` Python 2.7 backport. If you want to be able to work with S3, you'll need `boto3` (and a valid and properly set up AWS account).

To run the tests, you'll need the brilliant `moto` library, the `backports.tempfile` library for Python 2.7 compatibility, and `Keras` and `sckit-learn`.

For logging, simply grab the Velox logger by the `velox` handle.

You can install Velox using `pip install velox`. For more detailed info, visit our [website](https://vaitech.io/Velox).

## Basic Usage
---

For 90% of use cases, the `velox.lite` submodule provides a lightweight version of Velox's binary 
management capabilities.

The core functionality is geared towards continuous deployment environments in
the machine learning world, where consistency and versioning of binary objects
is key to maintain system integrity.

Suppose we build a simple [`scikit-learn`](http://scikit-learn.org/) model that we want to be available somewhere else in a secure, verifyable manner. 

Suppose a data scientist trains the following model.
<!--begin_code-->

    #!python
    import os

    from sklearn.linear_model import SGDClassifier
    from sklearn.datasets import make_blobs
    from velox.lite import save_object

    X, y = make_blobs()

    clf = SGDClassifier()
    clf.fit(X, y)

    save_object(
        obj=clf, 
        name='CustomerModel', 
        prefix='s3://myprodbucket/ml/models', 
        secret=os.environ.get('ML_MODELS_SECRET')
    )
<!--end_code-->

Elsewhere, on a production server, we could easily load this model and
verify it's integrity.

<!--begin_code-->

    #!python
    import os

    from velox.lite import load_object
    try:
        clf = load_object(
            name='CustomerModel', 
            prefix='s3://myprodbucket/ml/models', 
            secret=os.environ.get('ML_MODELS_SECRET')
        )
    except RuntimeError:
        raise RuntimeError('Invalid Secret!')

    # do things with clf...
<!--end_code-->

For more advanced and fine-grained nuanced versioning and constraint satisfaction, read the following section regarding the `velox.obj.VeloxObject`

## `VeloxObject` Abstract Base Class
---

Functionality is exposed using the ``velox.obj.VeloxObject`` abstract base class (ABC). A subclass of a `velox.obj.VeloxObject` needs to implement three things in order for the library to know how to manage it.

* Your class must be defined with a `velox.obj.register_object` decorator around it.
* Your class must implement a `_save` object method that takes as input a file object and does whatever is needed to save the object.
* Your class must implement a `_load` class method (with the `@classmethod` decorator) that takes as input a file object and reconstructs and returns an instance of your class.

This allows you to abstract away much of the messiness in bookkeeping.

Here is a simple example showing all required components.

<!--begin_code-->

    #!python
    import dill
    from sklearn.linear_model import SGDClassifier
    from sklearn.pipeline import make_pipeline
    from velox import register_object, VeloxObject

    @register_object(
        registered_name='textclassifier',
        version='0.1.1-rc',
        version_constraints='>=0.1.0,<0.2.0'
    )
    class TextClf(VeloxObject):
        def __init__(self):
            super(TextClf, self).__init__()
            self._model = make_pipeline(TfidfVectorizer(), SGDClassifier())

        def _save(self, fileobject):
            dill.dump(self._model, fileobject)

        @classmethod
        def _load(cls, fileobject):
            obj = cls()
            setattr(obj, '_model', dill.load(fileobject))
            return obj

        def predict(self, texts):
            return self._model.predict(texts)

        def fit(self, texts, labels):
            return self._model.fit(texts, labels)
<!--end_code-->


Here is a full example using [`gensim`](https://github.com/RaRe-Technologies/gensim) to build a topic model and keep track of all the necessary ETL-type objects that follow:

<!--begin_code-->

    #!python


    import os
    import shutil
    import tarfile
    import tempfile

    from gensim.corpora import Dictionary
    from gensim.models.ldamulticore import LdaMulticore
    from spacy.en import English

    from velox import VeloxObject, register_object


    nlp = English()


    @register_object('lda', '0.2.1')
    class LDAModel(VeloxObject):

        def __init__(self, n_components=50):
            super(LDAModel, self).__init__()
            self._n_components = n_components

        @staticmethod
        def tokenize(text):
            return [
                t.text.lower()
                for t in nlp(unicode(text))
                if t.text.strip()
            ]

        def fit(self, texts, passes=5, n_workers=1,
                no_below=5, no_above=0.2):
            tokenized = map(self.tokenize, texts)

            self._dictionary = Dictionary(tokenized)
            self._dictionary.filter_extremes(
                no_below=no_below,
                no_above=no_above
            )

            self._lda = LdaMulticore(
                corpus=map(
                    self._dictionary.doc2bow,
                    tokenized
                ),
                workers=n_workers,
                num_topics=self._n_components,
                id2word=self._dictionary,
                passes=passes
            )

            return self

        def transform(self, texts):
            feats = (
                self._dictionary.doc2bow(self.tokenize(q))
                for q in texts
            )
            vecs = list(self._lda[feats])

            X = np.zeros((len(vecs), self._lda.num_topics))
            for i, v in enumerate(vecs):
                ix, val = zip(*v)
                X[i][np.array(ix)] = val
            return X

        def _save(self, fileobject):
            with tarfile.open(fileobject.name, 'w:') as tf:
                with tempfile.NamedTemporaryFile() as tmp:
                    self._dictionary.save(tmp.name)
                    tf.add(tmp.name, 'dict.model')

                tmpdir = tempfile.mkdtemp()

                sp = os.path.join(tmpdir, 'lda.model')
                self._lda.save(sp)

                tf.add(tmpdir, 'lda.dir')

        @classmethod
        def _load(cls, fileobject):

            tmpdir = tempfile.mkdtemp()
            with tarfile.open(fileobject.name, 'r:*') as tf:
                tf.extractall(tmpdir)

            model = cls()

            _dictionary = Dictionary.load(
                os.path.join(tmpdir, 'dict.model')
            )
            _lda = LdaMulticore.load(
                os.path.join(tmpdir, 'lda.dir', 'lda.model')
            )

            setattr(model, '_dictionary', _dictionary)
            setattr(model, '_lda', _lda)

            setattr(model, '_n_components', model._lda.num_topics)
            shutil.rmtree(tmpdir)

            return model

<!--end_code-->

Voilà! Now, let's say you have a list of strings, and you wanted to train this 
model:

<!--begin_code-->
    #!python

    texts = [...]

    lda = LDAModel(128)
    lda.fit(texts)

    T = lda.transform(texts)

    lda.save('s3://my-ci-bucket/models/foo')
<!--end_code-->

Elsewhere, (i.e., a production server, etc.) you can load the latest model like
so:

<!--begin_code-->

    #!python
    production_lda = LDAModel.load('s3://my-ci-bucket/models/foo')
    T = production_lda.transform(...)
<!--end_code-->

In many environments, we would like the model to get hotswapped when a new model 
is uploaded to `s3`. Velox makes this easy! As long as the model stays in 
memory, we can use a async reload thread to poll the `prefix` location for
updated models!

<!--begin_code-->

    #!python

    production_lda = LDAModel.load('s3://my-ci-bucket/models/foo')
    production_lda.reload(
        prefix='s3://my-ci-bucket/models/foo', 
        scheduled=True, 
        minutes=5
    )



    T = production_lda.transform(...)
<!--end_code-->

Velox manages many components of principled serialization and blob versioning / management in a framework agnostic way. Our vision is to provide a simple, common interface for models from any framework, be it PyTorch, Keras, TensorFlow, Scikit-learn, or any other.
"""
import logging

# supress "No handlers could be found for logger XXXXX" error
logging.getLogger('velox').addHandler(logging.NullHandler())

__version__ = '0.5.0'

from .obj import (VeloxObject, register_object, load_velox_object,
                  register_object)

from . import filesystem
from . import exceptions
from . import tools
from . import obj
from . import wrapper
from . import lite

__all__ = ['filesystem', 'exceptions', 'tools', 'obj', 'wrapper', 'lite']
